import os
import sys
from selectors import BaseSelector
from typing import Any, cast
from argparse import ArgumentTypeError
from .pipeline import pipeline
from .protocol import Message, send_to_moulti
Args = dict[str, Any]

def pint(value: str) -> int:
	integer_value = int(value)
	if integer_value < 0:
		raise ArgumentTypeError('expected a positive integer')
	return integer_value

def float_str(value: str) -> str:
	_ = float(value)
	return value

def bool_or_int(value: str) -> bool|int:
	if value.lower() == 'false':
		return False
	if value.lower() == 'true':
		return True
	return int(value)

def handle_reply(reply: Message) -> None:
	success = reply.get('done') is True
	if not success:
		fallback = 'alas, no error message was provided.'
		error_message = reply.get('error', fallback)
		sys.stderr.write(f'Something went wrong: {error_message}\n')
	sys.exit(0 if success else 1)

def send_to_moulti_and_handle_reply(message: Message) -> None:
	reply = cast(dict[str, Any], send_to_moulti(message))
	handle_reply(reply)

def send_no_none_to_moulti_and_handle_reply(args: Args) -> None:
	no_none_args = {k:v for (k,v) in args.items() if v is not None}
	send_to_moulti_and_handle_reply(no_none_args)

def send_delete(args: Args) -> None:
	errors = pipeline((step_id, {**args, 'id': step_id}, None) for step_id in args['id'])
	sys.exit(1 if errors else 0)

def call_all(calls: list[Any]) -> None:
	for call in calls:
		call[0](*call[1:])

def clean_selector(selector: BaseSelector|None, close_fds: bool = False, close: bool = True) -> None:
	"""
	Clean a Selector object, as offered by the selectors module, by systematically unregistering and optionally
	closing all registered file descriptors then closing the Selector object itself.
	"""
	if not selector:
		return
	for selectorkey in list(selector.get_map().values()):
		selector.unregister(selectorkey.fileobj)
		if close_fds:
			if hasattr(selectorkey.fileobj, 'close'):
				selectorkey.fileobj.close()
			elif isinstance(selectorkey.fileobj, int):
				os.close(selectorkey.fileobj)
	if close:
		selector.close()

DEFAULT_TAB_SIZE = 8

def get_tab_size() -> int:
	try:
		return pint(os.environ.get('MOULTI_TAB_SIZE', ''))
	except Exception:
		return DEFAULT_TAB_SIZE

TAB_SIZE = get_tab_size()
TAB_SPACES_STR = ' '*TAB_SIZE
TAB_SPACES_BYTES = b' '*TAB_SIZE

ANSI_ESCAPE_SEQUENCE_STR = '\x1b'
ANSI_ESCAPE_SEQUENCE_BYTES = b'\x1b'

ANSI_RESET_SEQUENCES_STR = ('\x1b[0m', '\x1b[m')
ANSI_RESET_SEQUENCES_BYTES = (b'\x1b[0m', b'\x1b[m')



import os
import sys
from typing import Any, Generator
from .protocol import moulti_connect, send_json_message, recv_json_message

def pipeline(messages: Generator, read_size: int = 1024**2) -> int:
	"""
	Pipeline multiple messages to a Moulti instance.
	messages should be a generator that yields triplets:
	- step id (str; None for instance properties)
	- Moulti message to send (dict[str, Any])
	- file descriptor number (int) if it is necessary to pass data, None otherwise
	read_size is used for all "pass" operations.
	Note: so far, this is the only function in Moulti that pipelines messages and replies. Therefore, it is also the
	only function that actually leverages message ids.
	"""
	errors = 0
	sent_messages: dict[str, bool] = {}
	def send_message(*args: Any) -> None:
		msg_id = send_json_message(*args)
		sent_messages[msg_id] = False
	filenos = []
	with moulti_connect() as moulti_socket:
		for step_id, data, fileno in messages:
			# Send all messages using protocol pipelining:
			try:
				# Properties, typically a command that creates and configures a step:
				send_message(moulti_socket, data)
				# Contents:
				if fileno is not None:
					filenos.append(fileno)
					pass_msg = {'command': 'pass', 'id': step_id, 'read_size': read_size}
					send_message(moulti_socket, pass_msg, [fileno])
			except Exception as exc:
				errors += 1
				sys.stderr.write(f'Error with {step_id}/{data}/{fileno}: {exc}\n')
		# Receive replies to all sent messages:
		while not all(sent_messages.values()):
			reply, _ = recv_json_message(moulti_socket, 0)
			if 'msgid' in reply and reply['msgid'] in sent_messages:
				sent_messages[reply['msgid']] = True
				if reply.get('done') is not True:
					errors += 1
					sys.stderr.write(f'Error: {reply.get("error")}\n')
		for fileno in filenos:
			os.close(fileno)
	return errors
